﻿------------------------------------------------------------------------------
-- $Id$
------------------------------------------------------------------------------
--                            clcInfo_ArmsWarrior.lua
--
--    Copyright (C) 2014 jlongwell
--
--  This program is free software: you can redistribute it and/or modify
--  it under the terms of the GNU Lesser General Public License as published
--  by the Free Software Foundation, either version 3 of the License, or
--  (at your option) any later version.
--
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--  GNU General Public License for more details.
--
--  You should have received a copy of the GNU Lesser General Public License
--  along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
--  This file is for use with clcInfo by abija.
--
--  abija has distributed clcInfo All Rights Reserved.  For more information
--  go to: http://wow.curseforge.com/addons/clctracker/
--
------------------------------------------------------------------------------
-- Author:
--      jlongwell (Andërs@Sargeras)
------------------------------------------------------------------------------
-- Description:
--   Class Module implementation for clcInfo_ArmsWarrior. 
------------------------------------------------------------------------------
if not clcInfo then return end       -- need clcInfo to run
if clcInfo_Warrior then return end   -- don't load if already loaded
local ENGLISH_CLASS_NAME = "WARRIOR"
local _, englishClass  = UnitClass("player") -- only load for warriors
if englishClass ~= ENGLISH_CLASS_NAME then return end

-----------------NAMESPACE---------------------------------------
clcInfo_Warrior = {}
clcInfo_Warrior.VERSION = 2.0
clcInfo_Warrior.CLASS = ENGLISH_CLASS_NAME

-----------------WARRIOR SPELLS----------------------------------
clcInfo_Warrior.DELAY = 10
clcInfo_Warrior.DISCOUNT = 100

clcInfo_Warrior.spells = {
    --Arms
    ["Enrage"]              = 57519,
    ["Mortal Strike"]       = 12294,
	["Bloodthirst"]       	= 23881,
	["Raging Blow"]         = 85288,
	["Wild Strike"]         = 100130,
	["Whirlwind"]			= 1680,
	["Rend"]				= 772,
	["Dragon Roar"]			= 118000,
	["Storm Bolt"]			= 107570,
	["Ravager"]				= 152277,
    --Generic Warrior
    ["Battle Shout"]        = 6673,
    ["Berserker Rage"]      = 18499,
    ["Colossus Smash"]      = 167105,
    ["Commanding Shout"]    = 469,
    ["Demoralizing Shout"]  = 1160,
    ["Execute"]             = 5308,
    ["Recklessness"]        = 1719,
    ["Thunder Clap"]        = 6343,

}
--localize spells
local spell_name
for key, value in pairs(clcInfo_Warrior.spells) do
    spell_name = GetSpellInfo(value)
    if not spell_name then spell_name = "Unknown Spell" end
    clcInfo_Warrior.spells[key] = { id = value, name = spell_name }
end

--Generic Warrior cds for all specs
clcInfo_Warrior.basic_cds = {
    ["Battle Shout"]        =clcInfo_Warrior.spells["Battle Shout"].id,
    ["Challenging Shout"]   =1161,
    ["Charge"]              =100,
    ["Commanding Shout"]    =clcInfo_Warrior.spells["Commanding Shout"].id,
    ["Disarm"]              =676,
    ["Enraged Regeneration"]=55694,
    ["Hamstring"]           =1715,
    ["Heroic Leap"]         =6544,
    ["Heroic Throw"]        =57755,
    ["Intercept"]           =20252,
    ["Intervene"]           =3411,
    ["Intimidating Shout"]  =5246,
    ["Pummel"]              =6552,
    ["Rallying Cry"]        =97462,
    ["Recklessness"]        =1719,
    ["Retaliation"]         =20230,
    ["Shattering Throw"]    =64382,
    ["Shield Wall"]         =871,
    ["Spell Reflection"]    =23920,
    ["Thunder Clap"]        =clcInfo_Warrior.spells["Thunder Clap"].id,
    ["Victory Rush"]        =34428,
    ["Whirlwind"]           =1680,
}

-----------------EVENTS------------------------------------------
-- event handling.
--clcInfo_Warrior.EventFrame = CreateFrame("Frame", "clcInfo_WarriorFrame",
--                                           UIParent)
--clcInfo_Warrior.EventFrame:Hide()
--clcInfo_Warrior.EventFrame:SetScript("OnEvent", function(self, event, ...)
--    clcInfo_Warrior[event](clcInfo_Warrior, event, ...)
--end)
-- To add an event, write a clcInfo_Warrior function to
-- handle it and then register the event.  For instance:
--[[
function clcInfo_Warrior.A_WOW_EVENT(self, event, ...)
    -- handle the event
end
clcInfo_Warrior.EventFrame:RegisterEvent("A_WOW_EVENT")
]]--


-----------------FUNCTIONS---------------------------------------
-- Configuration functions - performance not an issue
--

-- creates the mod name. Each spec needs its own module
function clcInfo_Warrior.make_modName(specname)
    return "____warrior_" .. specname
end

-- Each spec needs its own priority list of actions as well.
function clcInfo_Warrior.StoreQueue(spec, q)
    clcInfo_Warrior[spec .. "_q"] = q
end

function clcInfo_Warrior.GetQueue(spec)
    return clcInfo_Warrior[spec .. "_q"]
end

-- Create the class module table with the name in class_module_name
-- Return the module and the module name.
function clcInfo_Warrior:init_module(specname)
    local class_module_name = self.make_modName(specname)
    local class_module = clcInfo:RegisterClassModule(class_module_name)
    return class_module, class_module_name
end

-- update the actions based on the prio settings the user provides
-- set the queue for the given spec
function clcInfo_Warrior.UpdateQueue(actions, tdb, spec)
    local q = {}
    local check = {} -- used to check for duplicates
    for key, value in ipairs(tdb.priority) do
        if not check[value] then
            if value ~= "none" then
                if actions[value] then
                    table.insert(q, value)
                --else
                    --print("clcInfo_Warrior invalid action:" .. value)
                end
            end
        else
            print("Ignoring duplicate action:"..key..")"..value)
        end
    end
    clcInfo_Warrior.StoreQueue(spec, q) 
end

-- register module and use default settings if version has changed
-- returns the template db, the queue, and the get and set functions for the
-- module options screen
-- called by clcInfo when templates update
function clcInfo_Warrior:OnTemplatesUpdate(class_module_name, actions,
                                         tdb_default, spec)
    local tdb = clcInfo:RegisterClassModuleTDB(class_module_name, tdb_default)
    local GetTDB, SetTDB, GetPriority, SetPriority
    if tdb then
        if not tdb.version or tdb.version < self.VERSION then
            -- fix stuff
            clcInfo.AdaptConfigAndClean(class_module_name, tdb, tdb_default)
            tdb.version = self.VERSION
        end

        self.UpdateQueue(actions, tdb, spec)
        GetTDB = clcInfo_Warrior.newGetTDB(tdb)
        SetTDB = clcInfo_Warrior.newSetTDB(tdb)
        GetPriority = clcInfo_Warrior.newGetPriority(tdb)
        SetPriority = clcInfo_Warrior.newSetPriority(tdb, actions, spec)
    end

    return tdb, GetTDB, SetTDB, GetPriority, SetPriority
end

-- called by clcInfo when initializing
-- returns the character db and the get and set functions for the module
-- options screen
function clcInfo_Warrior:OnInitialize(class_module_name, db_default)
    local db = clcInfo:RegisterClassModuleDB(class_module_name, db_default)
    local GetDB = self.newGetDB(db)
    local SetDB = self.newSetDB(db)
    return db, GetDB, SetDB
end

-- initializes the template db and q based on command line arguments
function clcInfo_Warrior:CmdLine(tdb, args, actions, spec)
    -- add args to options
    local num = 0
    local MAX_PRIORITY = 0
    for key, action in pairs(actions) do
        MAX_PRIORITY = MAX_PRIORITY + 1
    end

    for i, arg in ipairs(args) do
        if actions[arg] then
            if num < MAX_PRIORITY then
                num = num + 1
                tdb.priority[num] = arg
            else
                print("clcInfo_Warrior too many cmdLine priorities.\nMax is:"
                      .. MAX_PRIORITY)
            end
        else -- unrecognized action
            print(arg .. " not a recognized action.  Ignored.")
        end
    end
    
    -- none on the rest
    for i = num + 1, MAX_PRIORITY do
        db.priority[i] = "none"
    end

    -- redo queue
    self.UpdateQueue(actions, tdb, spec)
    clcInfo:UpdateOptions()
end

---FUNCTIONS: Frequently called-------------------------------------------------
-- Ability Icon Functions -> keep these fast!
local GetTime = GetTime
local GetSpellCooldown = GetSpellCooldown
local UnitPower = UnitPower
local UnitAura = UnitAura
local GetQueue = clcInfo_Warrior.GetQueue
local clcInfo_debug = clcInfo.debug
local emod = clcInfo.env
local IconSpell = clcInfo.env.IconSpell
local gcdSpellId = clcInfo_Warrior.spells["Whirlwind"].id

-- locals for keeping track of time
local current_time
local gcd_duration = 0

-- casting_name (basically for slam), current ability id, next ability id
local casting_name, cur_id, next_id

-- rage
local rage

-- gets the initial status: time, gcd situation
local function SetStatus()
    cur_id = nil

    -- current time
    current_time = GetTime()
    
    local start, gcd = GetSpellCooldown(gcdSpellId)
    if gcd > 0 then
        gcd_duration = gcd
    end

    local end_time, gcd_left
    local casting_left = 0
    -- cast_left will either be the cast time left on casting_name or
    -- the time left in the current gcd.
    casting_name, _, _, _, _, end_time = UnitCastingInfo("player")

    if casting_name then -- channelling a spell    
        casting_left = end_time/1000 - current_time
    end
      gcd_left = start + gcd - current_time
    if gcd_left < 0 then gcd_left = 0 end

    -- move time ahead to when user finishes current cast
    current_time = current_time + max(casting_left, gcd_left)   

    rage = UnitPower("player", SPELL_POWER_RAGE)
end

-- given a buff or debuff, the unit, and a filter, returns the time left based
-- on the current time and time left on the gcd
-- returns 0 if no such buff or debuff is present
function clcInfo_Warrior.GetTimeLeft(unit, spell, filter)
    local time_left = 0
    local name, expires
    name, _, _, _, _, _, expires = UnitAura(unit, spell, nil, filter)
    if name then
        time_left = expires - current_time
    end
    return time_left
end

-- given a spell id, returns the cd left on it based on the current time
-- variables
function clcInfo_Warrior.GetCooldown(id)
    local start, duration = GetSpellCooldown(id)
    local cd = start + duration - current_time
    if cd < 0 then return 0 end
    return cd
end

-- Calculates the target's health as a percentage.
local UnitHealth = UnitHealth
local UnitHealthMax = UnitHealthMax
function clcInfo_Warrior.CalculateTargetHealthPct()
    if not UnitExists("target") then return 100 end
    return floor(UnitHealth("target") / UnitHealthMax("target") * 100)
end

-- Decides which shout to cast
local bshoutId = clcInfo_Warrior.spells["Battle Shout"].id
local cshoutId = clcInfo_Warrior.spells["Commanding Shout"].id
local BattleShout = clcInfo_Warrior.spells["Battle Shout"].name
local CommandingShout = clcInfo_Warrior.spells["Commanding Shout"].name

function clcInfo_Warrior.PickShout()
    --if I cast battle shout, cast it again
    if (UnitAura("player", BattleShout, nil, "HELPFUL|PLAYER")) then
        return bshoutId
    end
    
    --if I cast commanding shout, cast it again
    if (UnitAura("player", CommandingShout, nil, "HELPFUL|PLAYER")) then
        return cshoutId
    end
    
    return bshoutId --everything is covered so just cast battle
end

-- figures out which ability should be used next
local function GetNextAction(actions, q)
    local n = #q
    local action_index

    -- parse once, get cooldowns, return first 0
    for i = 1, n do
        action_index = q[i]
        local action = actions[action_index]
        local cd = action.GetCD(cur_id, rage)
        if clcInfo_debug.enabled then
            clcInfo_debug:AddBoth(action_index, cd)
        end
        if cd == 0 then
            return action.id(), action_index
        end
        action.cd = cd
    end
    
    -- parse again, return min cooldown
    local minQ = 1
    local minCd = actions[q[1]].cd
    for i = 2, n do
        local action_cd = actions[q[i]].cd
        if minCd > action_cd then
            minCd = action_cd
            minQ = i
        end
    end
    action_index = q[minQ]
    return actions[action_index].id(), action_index
end

-- determines the current and next abilities
local executeId = clcInfo_Warrior.spells["Execute"].id
local function WarriorRotation(actions, q)
    SetStatus()
    
    local action
    cur_id, action = GetNextAction(actions, q)
clcInfo_debug.enabled = false
    -- Adjust state variables based on the action suggestion
    local _, _, _, cost, _, _, casting_time = GetSpellInfo(cur_id)
    current_time = current_time + max(gcd_duration, casting_time/1000)
    
    --one day get fancy and calculate regen using swing timer
    --for now do this
    local rage_adjust = 0
    if cur_id == executeId and rage >= 10 then
        rage_adjust = -min(rage - 10, 20)
    elseif cur_id == bshoutId or cur_id == cshoutId then
        rage_adjust = 20
    end
    rage = rage - cost + rage_adjust + (gcd_duration * ((1/3) + 6.5))
    next_id, action = GetNextAction(actions, q)
end

local nextSkill
local devName = clcInfo_Warrior.spells["Whirlwind"].name
local local_range_per_skill = nil

-- Get current ability
function clcInfo_Warrior.IconCurrent(actions, spec, rangePerSkill)
    local_range_per_skill = rangePerSkill
    local q = GetQueue(spec)
    WarriorRotation(actions, q)
    if nextSkill then nextSkill:DoUpdate() end
    return IconSpell(cur_id, local_range_per_skill or devName)
end

-- Determine Next ability
local function NextExec()
    return IconSpell(next_id, local_range_per_skill or devName)
end

-- reset next ability
local function ExecCleanupNext()
    nextSkill = nil
end

-- Get Next ability
function clcInfo_Warrior.IconNext()
    nextSkill = emod.___e
    nextSkill:SetScript("OnUpdate", nil)
    nextSkill.exec = NextExec
    nextSkill.ExecCleanup = ExecCleanupNext
end